/*
 * Copyright (c) 2016 MariaDB Corporation Ab
 * Copyright (c) 2023 MariaDB plc, Finnish Branch
 *
 * Use of this software is governed by the Business Source License included
 * in the LICENSE.TXT file and at www.mariadb.com/bsl11.
 *
 * Change Date: 2024-06-03
 *
 * On the date above, in accordance with the Business Source License, use
 * of this software will be governed by version 2 or later of the General
 * Public License.
 */

#include <iostream>
#include <algorithm>
#include <maxbase/assert.h>
#include <maxbase/maxbase.hh>
#include <maxbase/worker.hh>

using namespace maxbase;
using namespace std;

namespace
{

// TODO: Put this in some common place.
int64_t get_monotonic_time_ms()
{
    struct timespec ts;
    MXB_AT_DEBUG(int rv = ) clock_gettime(CLOCK_MONOTONIC, &ts);
    mxb_assert(rv == 0);

    return ts.tv_sec * 1000 + ts.tv_nsec / 1000000;
}

class TimerTest
{
public:
    static int s_ticks;

    TimerTest(Worker* pWorker, int* pRv, int32_t delay)
        : m_id(s_id++)
        , m_worker(*pWorker)
        , m_delay(delay)
        , m_at(get_monotonic_time_ms() + delay)
        , m_rv(*pRv)
    {
    }

    ~TimerTest()
    {
        std::sort(m_tick_durations.begin(), m_tick_durations.end());
        size_t idx = (m_tick_durations.size() - 1) * 0.95;
        int64_t diff = m_tick_durations[idx];

        cout << "Delay: " << delay() << "ms, "
             << "95th percentile: " << diff << "ms, "
             << "Ticks: " << m_tick_durations.size() << ", "
             << "95th percentile index: " << idx
             << endl;

        if (diff > 50)
        {
            cout << "Error: 95th percentile difference between expected and happened > 50: " << diff << endl;
            m_rv = EXIT_FAILURE;
        }
    }

    int32_t delay() const
    {
        return m_delay;
    }

    bool tick(Worker::Call::action_t action)
    {
        bool rv = false;

        if (action == Worker::Call::EXECUTE)
        {
            int64_t now = get_monotonic_time_ms();
            int64_t diff = abs(now - m_at);
            m_tick_durations.push_back(diff);
            cout << m_id << ": " << diff << endl;

            m_at += m_delay;

            if (--s_ticks < 0)
            {
                m_worker.shutdown();
            }

            rv = true;
        }

        return rv;
    }

private:
    static int s_id;

    int     m_id;
    Worker& m_worker;
    int32_t m_delay;
    int64_t m_at;
    int&    m_rv;

    std::vector<int64_t> m_tick_durations;
};

int TimerTest::s_id = 1;
int TimerTest::s_ticks;

int run()
{
    int rv = EXIT_SUCCESS;

    TimerTest::s_ticks = 100;

    Worker w;

    TimerTest t1(&w, &rv, 200);
    TimerTest t2(&w, &rv, 300);
    TimerTest t3(&w, &rv, 400);
    TimerTest t4(&w, &rv, 500);
    TimerTest t5(&w, &rv, 600);

    w.execute([&]() {
        w.dcall(t1.delay(), &TimerTest::tick, &t1);
        w.dcall(t2.delay(), &TimerTest::tick, &t2);
        w.dcall(t3.delay(), &TimerTest::tick, &t3);
        w.dcall(t4.delay(), &TimerTest::tick, &t4);
        w.dcall(t5.delay(), &TimerTest::tick, &t5);

        // MXS-5248: Check that dcalls being processed get removed by cancel_dcall.
        std::vector<Worker::DCId> ids;

        for (int i = 0; i < 100; i++)
        {
            ids.push_back(w.dcall(1ms, [](){
                return true;
            }));
        }

        ids.push_back(w.dcall(1ms, [&w, ids](){
            for (auto id : ids)
            {
                w.cancel_dcall(id);
            }

            return false;
        }));
    }, mxb::Worker::EXECUTE_QUEUED);

    w.run();

    return EXIT_SUCCESS;
}
}

int main()
{
    mxb::MaxBase mxb(MXB_LOG_TARGET_STDOUT);

    return run();
}
