/*
 * Copyright (c) 2016 MariaDB Corporation Ab
 * Copyright (c) 2023 MariaDB plc, Finnish Branch
 *
 * Use of this software is governed by the Business Source License included
 * in the LICENSE.TXT file and at www.mariadb.com/bsl11.
 *
 * Change Date: 2024-06-03
 *
 * On the date above, in accordance with the Business Source License, use
 * of this software will be governed by version 2 or later of the General
 * Public License.
 */

#include <maxtest/testconnections.hh>
#include <string>
#include <maxbase/format.hh>
#include "mariadbmon_utils.hh"

using std::string;
using mxt::MaxScale;
using cooperative_monitoring::MonitorInfo;

namespace
{
// The test runs two MaxScales with two monitors each.
namespace MonitorID
{
const int UNKNOWN = -1;
const int ONE_A = 1;
const int ONE_B = 2;
const int TWO_A = 3;
const int TWO_B = 4;
}

MonitorInfo monitors[] = {
    {MonitorID::ONE_A,   "MariaDB-Monitor1A"},
    {MonitorID::ONE_B,   "MariaDB-Monitor1B"},
    {MonitorID::TWO_A,   "MariaDB-Monitor2A"},
    {MonitorID::TWO_B,   "MariaDB-Monitor2B"},
    {MonitorID::UNKNOWN, "none"             },
};

const int failover_mon_ticks = 6;
const int mxs_switch_ticks = 6;

const MonitorInfo* get_primary_monitor(TestConnections& test);
void test_failover(TestConnections& test, MaxScale& maxscale);
bool release_monitor_locks(TestConnections& test, const MonitorInfo& mon_info);
void ensure_primary_monitor(TestConnections& test, const MonitorInfo& primary_monitor);

void test_main(TestConnections& test)
{
    test.expect(test.n_maxscales() >= 2, "At least 2 MaxScales are needed for this test. Exiting");
    if (!test.ok())
    {
        return;
    }

    auto& mxs1 = *test.maxscale;
    auto& mxs2 = *test.maxscale2;
    monitors[0].maxscale = &mxs1;
    monitors[1].maxscale = &mxs1;
    monitors[2].maxscale = &mxs2;
    monitors[3].maxscale = &mxs2;

    test.tprintf("Starting MaxScales. MariaDB-Monitor1A should acquire server locks.");
    auto* primary_mon1 = &monitors[0];
    ensure_primary_monitor(test, *primary_mon1);

    mxs1.check_print_servers_status(mxt::ServersInfo::default_repl_states());
    mxs2.check_print_servers_status(mxt::ServersInfo::default_repl_states());

    if (test.ok())
    {
        // Test a normal failover.
        test_failover(test, *primary_mon1->maxscale);
    }

    // If ok so far, stop the MaxScale with the current primary monitor.
    if (test.ok())
    {
        auto* previous_primary_maxscale = primary_mon1->maxscale;
        test.tprintf("Stopping %s.", previous_primary_maxscale->node_name().c_str());
        previous_primary_maxscale->stop();
        MaxScale* expect_primary_maxscale = (previous_primary_maxscale == &mxs1) ? &mxs2 : &mxs1;
        // When swapping from one MaxScale to another, only waiting for monitor does not seem to be
        // 100% reliable. 1s sleep seems to ensure the switch has happened. A possible reason is that there
        // is some lag between a connection releasing a lock and that lock becoming available for other
        // connections to take.
        sleep(1);
        expect_primary_maxscale->wait_for_monitor(mxs_switch_ticks);
        const auto* primary_mon2 = get_primary_monitor(test);
        if (test.ok())
        {
            auto* current_primary_maxscale = primary_mon2->maxscale;
            test.expect(primary_mon2 != primary_mon1, "Primary monitor did not change.");
            test.expect(current_primary_maxscale == expect_primary_maxscale,
                        "Unexpected primary '%s'.", current_primary_maxscale->node_name().c_str());

            // Again, check that failover works. Wait a few more intervals since failover is not
            // immediately enabled on primary MaxScale switch.
            current_primary_maxscale->wait_for_monitor(failover_mon_ticks);
            test_failover(test, *current_primary_maxscale);
        }
        test.tprintf("Starting %s.", previous_primary_maxscale->node_name().c_str());
        previous_primary_maxscale->start();
        expect_primary_maxscale->wait_for_monitor(mxs_switch_ticks);
    }

    // If ok so far, do a rolling sweep through all four monitors by having each monitor release its
    // locks in turn.
    if (test.ok())
    {
        const char revisited[] = "Revisited the same monitor";
        test.tprintf("Testing rolling monitor swapping.");
        std::set<int> visited_monitors;
        while (visited_monitors.size() < 3 && test.ok())
        {
            const auto* primary_mon = get_primary_monitor(test);
            if (test.ok())
            {
                auto mon_id = primary_mon->id;
                test.expect(visited_monitors.count(mon_id) == 0, revisited);
                bool released = release_monitor_locks(test, *primary_mon);
                test.expect(released, "Releasing monitor locks failed");
                if (released)
                {
                    visited_monitors.insert(mon_id);
                    // The 'wait_for_monitor'-function causes the target monitor to tick faster than usual.
                    // This can cause issues when two separate MaxScales are involved, not leaving enough
                    // time for the next MaxScale to tick. Simply wait on both MaxScales.
                    sleep(1);
                    mxs1.wait_for_monitor(mxs_switch_ticks);
                    mxs2.wait_for_monitor(mxs_switch_ticks);
                }
            }
        }

        // Should have one monitor left.
        const auto* primary_mon = get_primary_monitor(test);
        if (test.ok())
        {
            test.expect(visited_monitors.count(primary_mon->id) == 0, revisited);
        }
    }

    if (test.ok())
    {
        // MXS-4902
        test.tprintf("Test that all manual commands fail on a monitor which does not have locks. Restart "
                     "both MaxScales to clear any lock release timeouts.");
        auto* primary_mon = &monitors[2];
        ensure_primary_monitor(test, *primary_mon);

        if (test.ok())
        {
            auto try_manual_command = [&](const char* cmdname, const char* opts) {
                string total_cmd = mxb::string_printf("call command mariadbmon %s MariaDB-Monitor1A %s 2>&1",
                                                      cmdname, opts);
                test.tprintf("Testing MaxCtrl command '%s'.", total_cmd.c_str());
                auto res = monitors[0].maxscale->maxctrl(total_cmd);
                const int expected_rc = 1;
                test.expect(res.rc == expected_rc, "Command '%s' returned %i when %i was expected.",
                            total_cmd.c_str(), res.rc, expected_rc);
                bool msg_found =
                    res.output.find("this MaxScale does not have exclusive locks") != string::npos;
                test.expect(msg_found, "Command output did not contain expected phrase. Output: '%s'",
                            res.output.c_str());
            };
            try_manual_command("switchover", "");
            try_manual_command("failover", "");
            try_manual_command("rejoin", "server4A");
            try_manual_command("reset-replication", "server1A");

            if (test.ok())
            {
                test.tprintf("Test a manual command on the primary monitor, it should succeed.");
                string cmd = mxb::string_printf("call command mariadbmon switchover %s",
                                                primary_mon->name.c_str());
                auto res = primary_mon->maxscale->maxctrl(cmd);
                test.tprintf("Command '%s' returned '%s'.", cmd.c_str(), res.output.c_str());
                test.expect(res.rc == 0, "Command failed on primary monitor");
            }
        }
        else
        {
            test.add_failure("MariaDB-Monitor1A is not secondary, cannot continue.");
        }
    }

    if (test.ok())
    {
        test.tprintf("Test successful!");
    }
}

const cooperative_monitoring::MonitorInfo* get_primary_monitor(TestConnections& test)
{
    return cooperative_monitoring::get_primary_monitor(test, monitors);
}

bool release_monitor_locks(TestConnections& test, const MonitorInfo& mon_info)
{
    string cmd = "call command mariadbmon release-locks " + mon_info.name;
    auto res = mon_info.maxscale->maxctrl(cmd);
    bool success = res.rc == 0 && (res.output == "OK" || res.output == "\"OK\"");
    test.expect(success, "MaxCtrl command failed.");
    return success;
}

void ensure_primary_monitor(TestConnections& test, const MonitorInfo& primary_monitor)
{
    auto* mxs1 = test.maxscale;
    auto* mxs2 = test.maxscale2;
    mxs1->stop();
    mxs2->stop();
    auto* primary_mxs = primary_monitor.maxscale;
    auto* other_mxs = primary_mxs == mxs1 ? mxs2 : mxs1;

    primary_mxs->start();
    // Ensure the correct monitor gets locks by restarting monitors in order.
    const MonitorInfo* other_monitor = nullptr;
    for (int i = 0; monitors[i].maxscale; i++)
    {
        auto& mon_info = monitors[i];
        if (mon_info.maxscale == primary_mxs)
        {
            primary_mxs->maxctrl(mxb::string_printf("stop monitor %s", mon_info.name.c_str()));
            if (mon_info.id != primary_monitor.id)
            {
                other_monitor = &mon_info;
            }
        }
    }
    sleep(1);
    const char start_mon_fmt[] = "start monitor %s";
    primary_mxs->maxctrl(mxb::string_printf(start_mon_fmt, primary_monitor.name.c_str()));
    primary_mxs->sleep_and_wait_for_monitor(1, 1);
    primary_mxs->maxctrl(mxb::string_printf(start_mon_fmt, other_monitor->name.c_str()));
    primary_mxs->wait_for_monitor(1);

    other_mxs->start_maxscale();
    other_mxs->wait_for_monitor(1);

    auto primary_mon = get_primary_monitor(test);
    test.expect(primary_mon->id == primary_monitor.id, "Wrong primary monitor. Got %i, expected %i.",
                primary_mon->id, primary_monitor.id);
}

void test_failover(TestConnections& test, MaxScale& maxscale)
{
    // Test a normal failover.
    mxt::ServerInfo first_master = maxscale.get_servers().get_master();
    test.expect(first_master.server_id > 0, "No master at start of failover");
    if (test.ok())
    {
        test.tprintf("Stopping %s and waiting for failover.", first_master.name.c_str());
        int master_node = first_master.server_id - 1;
        test.repl->stop_node(master_node);
        maxscale.wait_for_monitor(failover_mon_ticks);
        mxt::ServerInfo second_master = maxscale.get_servers().get_master();
        test.expect(second_master.server_id > 0, "No master after failover");
        if (test.ok())
        {
            test.tprintf("%s is now master.", second_master.name.c_str());
            test.expect(first_master.server_id != second_master.server_id,
                        "Master did not change, failover did not happen.");
        }
        test.tprintf("Starting %s.", first_master.name.c_str());
        test.repl->start_node(master_node);
        maxscale.wait_for_monitor(failover_mon_ticks);      // wait for rejoin, assume it works
    }
}
}

int main(int argc, char* argv[])
{
    TestConnections test;
    TestConnections::skip_maxscale_start(true);
    return test.run_test(argc, argv, test_main);
}
